#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <ctype.h>
#include <unistd.h>
#include <stdint.h>
#include "resize_image.h"

#include "config.h"
#include "far.h"
#include "mbox.h"
#include "main.h"
#include "worker.h"
#include "master.h"
#include "util.h"
#include "ops.h"

static pthread_t workerthr;
static int worker_quit_flag,worker_run;
mbox_t *worker_mbox;

static inp_t *files[200];
static int nfiles;
static inp_t *done[200];
static int ndone;

//static pthread_attr_t wattr;

static int worker_cut, worker_rotate, 
           worker_ipagenum, worker_delta, worker_edgefill, worker_mask;

void worker_stop() { worker_run= 0; }
void worker_quit() { worker_run= 0; worker_quit_flag= 1; }

static int worker_probe()
{
  char **L;
  inp_t *P;
  int n,i,j;
  int ch;
  static char *ext[]= { ".pgm" , NULL };

  if (!CONFIG.G_tempdir || !CONFIG.tempdir || !CONFIG.tempdir[0]) return 0;
  L= list_files(CONFIG.tempdir);
  if (!L) return 0;
  if (!L[0]) return ifree(L,0);

  file_types(CONFIG.tempdir, L);
  for(n=0;L[n];n++) ;

  for(i=0;i<n;i++) if (L[i][0]!='F') { L[i]= L[n-1]; n--; i--; }
  if (n==0) return ifree(L,0);

  n= filter_files_byext(n,L,ext);
  n= filter_files_byhid(n,L,0);

  if (n==0) return ifree(L,0);

  ch= 0;
  for(i=0;i<n;i++) 
  {
    for(j=0;j<nfiles;j++) if (!strcmp(files[j]->name,L[i]+1)) break;
    if (j!=nfiles) continue;
    for(j=0;j<ndone;j++) if (!strcmp(done[j]->name,L[i]+1)) break;
    if (j!=ndone) continue;
    P= calloc(1,sizeof(*P));
    P->name= strdup(L[i]+1);
    files[nfiles++]= P;
    ch= 1;
  }
  free(L);
  return ch;
}

static int fname_to_pgnum(char *name)
{
  int K;
  while(*name && !isdigit(*name)) name++;
  while(*name=='0') name++;
  if (!isdigit(*name)) K= 1;
                  else K= atoi(name);

  return (K-1)*worker_delta+worker_ipagenum;
}

static int process_file() {
  unsigned char *d, *td;
  unsigned int tw, th, w,h;
  unsigned int left,right,top,bottom;
  inp_t *P;
  int thr, have_thr;

  thr= CONFIG.threshold;
  if (CONFIG.G_threshold && thr>0 && thr<255) have_thr= 1;
                                         else have_thr= 0;

  if (!nfiles) return 0;
  P= files[nfiles-1];
  nfiles--;
  P->page_number= fname_to_pgnum(P->name);

  done[ndone++]= P; 

  if (read_image(P->name, &d, &w,&h)) return P->status= 'E';

  if (CONFIG.G_cutleft) left= CONFIG.cutleft; else left= 0;
  if (CONFIG.G_cutright) right= CONFIG.cutright; else right= 0;
  if (left<0) left= 0; if (right<0) right= 0;
  if (left || right)
  {
    sub_image(d, w,h, left,right, 0,0, &td, &tw, &th);
    free(d);
    d= td; w= tw; h= th;
  }
  if (worker_cut && have_thr && 
      !black_border(d,w,h,thr,&left,&right,&top,&bottom))
  {
    sub_image(d, w,h, left,right, 0,0, &td, &tw, &th);
    free(d);
    d= td; w= tw; h= th;
  }
  if (worker_rotate && have_thr && 
     !process_image(d,w,h,&td,&tw,&th,0,0,thr,
         P->page_number%2?'R':'L'))
  {
    free(d);
    d= td; w= tw; h= th;
  }
  if (worker_edgefill && have_thr)
    edge_fill(d, w,h, thr);
  if (worker_mask && CONFIG.G_mask && CONFIG.mask>0 && CONFIG.mask<255)
    apply_mask(d, w,h, CONFIG.mask);
  P->status= 'K';

  resize_image_g8_fixed(d,w,h, 400,500, &P->id, &P->iw, &P->ih);
  P->d= d; P->w= w; P->h= h;
  return 0;
}



void worker_start(int cut,int rot,int fil,int mask,int ipage,int delta)
{
  worker_run= 1;
  worker_cut= cut; worker_rotate= rot; 
  worker_delta= delta; worker_ipagenum= ipage;
  worker_edgefill= fil;
  worker_mask= mask;
  nfiles= ndone= 0;
  while(worker_run)
  {
    while (nfiles) process_file();
    if (!worker_probe()) usleep(500*1000);
  }
  while (worker_probe());
  while (nfiles) process_file();
  mbox_put(master_mbox,master_HANDLE_INPUT(done,ndone));
}

static void *worker_loop(void *dummy_arg)
{
  worker_quit_flag= 0;
  while(!worker_quit_flag)
  {
    worker_dispatch(mbox_get(worker_mbox));
  }
  mbox_put(master_mbox,master_HANDLE_DEAD_WORKER());
  return NULL;
}

void create_worker_thread()
{
  worker_mbox= mbox_new();
  pthread_create(&workerthr,NULL,worker_loop,NULL);
}

void join_worker_thread()
{
  pthread_join(workerthr, NULL);
}

#include "worker.c"
