#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include "pngio.h"
#include "mbox.h"
#include "far.h"
#include "main.h"
#include "writer.h"
#include "master.h"

static pthread_t writhr;
static int writer_quit_flag;
static mbox_t *writer_mbox;
static farchive_t *par;
static int writer_ok;

static void writer_quit()
{
  writer_quit_flag= 1;
}

static void writer_write_member
   (int page, unsigned int w, unsigned int h, unsigned char *data)
{
  unsigned char *pd; unsigned int pl;
  char buf[30];
  encode_png_g8(w,h,data, &pl, &pd);
  free(data);
  sprintf(buf,"page%04d.png",page);
  if (par) far_write_member(par,page,buf,pd,pl);
  free(pd);
}

static void writer_write_index
   (unsigned char *data)
{
  if (par)
    far_overwrite_member(par,0,data, 1000);
  free(data);
}

static void *writer_loop(void *dummy_arg)
{
  writer_quit_flag= 0;
  while(!writer_quit_flag)
  {
    writer_dispatch(mbox_get(writer_mbox));
  }
  mbox_put(master_mbox, master_HANDLE_DEAD_WRITER());
  return NULL;
}

void create_writer_thread(farchive_t *far)
{
   writer_ok= 0;
   if (!far) return ;
   mbox_destroy(&writer_mbox);
   writer_mbox= mbox_new();
   par= far;
   writer_ok= 1;
   pthread_create(&writhr, NULL, writer_loop, NULL);
}

#include "writer.c"

static int writer_done;

void master_handle_dead_writer()
{
  writer_done= 1;
}


void finish_writer_thread()
{
  if (!writer_ok) return ;
  writer_done= 0;
  mbox_put(writer_mbox, writer_QUIT());
  while(!writer_done) master_dispatch(mbox_get(master_mbox));
  pthread_join(writhr,NULL);
  writer_ok= 0;
}

void write_page(int page,unsigned int w,unsigned int h,unsigned char *d)
{
  if (!writer_ok) { free(d); return ; }
  mbox_put(writer_mbox, writer_WRITE_MEMBER(page,w,h,d));
}

void write_index(unsigned char *d)
{
  unsigned char *R;
  if (!writer_ok) return ;
  R= malloc(1000);
  memcpy(R,d,1000);
  mbox_put(writer_mbox, writer_WRITE_INDEX(R));
}


