#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include "gfx.h"
#include "far.h"
#include "config.h"
#include "ops.h"
#include "mbox.h"
#include "main.h"
#include "worker.h"
#include "master.h"
#include "util.h"

static char* file_name;
static farchive_t *par;
mbox_t *master_mbox;

typedef struct
{
  unsigned char pages[1000]; 
} parinfo_t;

static parinfo_t info;

static void reset_info() { memset(&info,0,sizeof(info)); }
static int write_info()  {
  if (!par) return 0;
  return far_write_member(par,0,"info", (unsigned char*) &info, sizeof(info));
}

typedef struct
{
  unsigned int w,h;
  unsigned char *data;

  unsigned int iw,ih;
  unsigned char *id;

  int x,y;

  int pagenum;
  char *fname;
} page_t;

static page_t pages[100];
static int n_pages;
static int selpage=-1;


static void mkpage(int idx,inp_t *S)
{
  page_t *P;

  P= pages+idx;
  if (P->id) free(P->id);
  if (P->data) free(P->data);
  memset(P,0,sizeof(P[0]));

  P->fname= S->name;
  P->pagenum= S->page_number;
  if (!S->d) mkerrpage(&P->id, &P->iw, &P->ih); 
     else { P->data= S->d; P->w= S->w; P->h= S->h;
            P->id= S->id; P->iw= S->iw; P->ih= S->ih; }
  free(S);
}

static int imax2(int a,int b) { return a>b? a:b; }
// static int imax3(int a,int b,int c) { return imax2(a,imax2(b,c)); }

static void print_num(int widget, int x,int y,int nr)
{
  char buf[20];
  int w,h; unsigned char *d;
  sprintf(buf,"%d",nr);
  gfx_text_data(1, buf,&w,&h,NULL,&d);
  gfx_canvas_put_graymap(Wpixlist,0x0000ff, x,y, w,h, d, w,h, 0,0);
  free(d);
}

static void layout_pages()
{
  int i;
  int DW, DH;
  int y= 30,x=30,maxy;

  maxy= 0;
  for(i=0;i<n_pages;i++)
  {
    if (i && i%2==0)
    {
      y= maxy+60;
      x= 30;
    }
    pages[i].y= y;
    maxy= imax2(y+pages[i].ih,maxy);
    pages[i].x= x;
    x+= pages[i].iw + 30;
  }
  DH= maxy+60;
  DW= 1008;
  gfx_canvas_setsize(Wpixlist, DW,DH);
  gfx_canvas_fill_rectangle(Wpixlist, 0xffFFff, 0,0,DW,DH);
  for(i=0;i<n_pages;i++)
  {
    gfx_canvas_put_graymap(Wpixlist,0x0000ff, pages[i].x, pages[i].y,
                           pages[i].iw, pages[i].ih,
                           pages[i].id, pages[i].iw, pages[i].ih,0,0);
    print_num(Wpixlist, pages[i].x+20, pages[i].y+pages[i].ih+10, 
                     pages[i].pagenum);
  }
}

/***********************************************************************
 OK. two buttons for backward and forward scanning. I will simply find
 and display the missing pages.
 ***********************************************************************/
static void compute_page_numbers(int *Rodd,int *Reven)
{
  int i;
  for(i=1;i<1001;i+=2) if (!info.pages[i]) break;
  if (i==1001) i= 0;
  *Rodd= i;
  for(i=998;i>2;i-=2) if (!info.pages[i]  && info.pages[i-1]) break;
  *Reven= i;
}

static int compar_scan(const void *a, const void *b)
{
  const inp_t *A,*B;
  A= *(const inp_t**) a;
  B= *(const inp_t**) b;
  return strcmp(A->name,B->name);
}

static void sort_inp(inp_t **s,int ns)
{
  qsort(s, ns, sizeof(s[0]), compar_scan);
}

static int worker_finished; // batch is finished
static int worker_done; // thread has exited
static inp_t **scans;
static int nscans;

void master_handle_input(inp_t **s,int n)
{
  scans= s;
  nscans= n;
  worker_finished= 1;
}

void master_handle_dead_worker()
{
  worker_done= 1;
}


static void scan_batch(int dir)
{
  int i;
  int odd, even;
  int start,delta;

  compute_page_numbers(&odd,&even);

  if (dir=='F')  { start= odd; delta= 2; }
          else   { start= even; delta= -2; }

  mbox_put(worker_mbox,worker_START(gfx_get_state(Wcut), 
                                    gfx_get_state(Wrotate),
                                    gfx_get_state(Wedge),
                                    gfx_get_state(Wmask),
                                    start,delta));
  scan_images();
  worker_stop();

  worker_finished= 0;
  while(!worker_finished) master_dispatch(mbox_get(master_mbox));

  sort_inp(scans, nscans);
  for(i=0;i<nscans;i++)
    mkpage(i, scans[i]);
  n_pages= nscans;
 
  layout_pages();
}

static void write_pages()
{
  int i;
  page_t *P;
  for(i=0;i<n_pages;i++)
  {
    P= pages+i;
    if (P->data) 
    {
      write_page(P->pagenum, P->w, P->h, P->data);
      P->data= NULL;
      info.pages[P->pagenum]= 1;
    }
    remove_image(P->fname);
  }
  write_index(info.pages);
  n_pages= 0;
}


static void get_config_str(int widget, char* flag, char **value)
{
  char *s;
  s= gfx_get_text(widget);
  if (!s || !s[0]) *flag= 0;
  else { *flag= 1; *value= s; }
}

static void show_config_int(int widget, int avail, int val)
{
  char buf[40];
  sprintf(buf,"%d",val);
  if (avail) gfx_set_text(widget,buf);
         else gfx_set_text(widget,"");
}

static void show_settings_window()
{
  gfx_event_t event;
  gfx_SCRsettings();
  if (CONFIG.G_scancmd) gfx_set_text(WsetSCAN, CONFIG.scancmd);
  if (CONFIG.G_tempdir) gfx_set_text(WsetTEMP, CONFIG.tempdir);
  show_config_int(WsetCLEFT, CONFIG.G_cutleft, CONFIG.cutleft);
  show_config_int(WsetCRIGHT, CONFIG.G_cutright, CONFIG.cutright);
  show_config_int(WsetTHR, CONFIG.G_threshold, CONFIG.threshold);
  show_config_int(WsetMASK, CONFIG.G_mask, CONFIG.mask);
  while(1) if (gfx_event(NULL, &event)) switch(event.type)
  {
  case WsetOK:
     get_config_str(WsetSCAN, &CONFIG.G_scancmd, &CONFIG.scancmd);
     get_config_str(WsetTEMP, &CONFIG.G_tempdir, &CONFIG.tempdir);
  case GFXEV_QUIT:
  case WsetCANCEL:
     gfx_leave_screen();
     return ;
  case WsetADJUST:
     get_config_str(WsetSCAN, &CONFIG.G_scancmd, &CONFIG.scancmd);
     get_config_str(WsetTEMP, &CONFIG.G_tempdir, &CONFIG.tempdir);
     show_image_params();
     show_config_int(WsetCLEFT, CONFIG.G_cutleft, CONFIG.cutleft);
     show_config_int(WsetCRIGHT, CONFIG.G_cutright, CONFIG.cutright);
     show_config_int(WsetTHR, CONFIG.G_threshold, CONFIG.threshold);
     show_config_int(WsetMASK, CONFIG.G_mask, CONFIG.mask);
     break;
  }
}

void test_scr(void (*f)())
{
  gfx_event_t event;
  (*f)();
  while(!gfx_event(NULL,&event) || event.type!=GFXEV_QUIT) ;
  gfx_leave_screen();
}


static void create_file()
{
  char *fn;

  static char *ext[]= { "Funky Archive (.far)", NULL };
  fn= gfx_select_file(1, NULL, ext);
  if (!fn) return ;
  if (file_name)
  {
    finish_writer_thread();
    if (par) far_close(&par);
    free(file_name);
    file_name= NULL;
  }
  if (far_create(fn,1000,&par))
  {
    gfx_ask("Error", "!OK",
        "error creating archive %s:\n%s",fn, far_error_message(par));
    far_close(&par);
    return ;
  }
  reset_info();
  if (write_info())
  {
    gfx_ask("Error", "!OK",
        "error creating archive %s:\n%s",fn, far_error_message(par));
    far_close(&par);
    return ;
  }
  file_name= fn;
  gfx_set_text(Wstatus, fn);
  create_writer_thread(par);
}

static void load_file()
{
  char *fn;
  char *name;
  unsigned char *d; size_t len;
  static char *ext[]= { "Funky Archive (.far)", NULL };
  fn= gfx_select_file(0, NULL, ext);
  if (!fn) return ;
  if (file_name)
  {
    finish_writer_thread();
    if (par) far_close(&par);
    free(file_name);
    file_name= NULL;
  }
  if (far_open(fn,&par))
  {
    gfx_ask("Error", "!OK",
        "error opening archive %s:\n%s",fn, far_error_message(par));
    far_close(&par);
    return ;
  }
  if (far_read_member(par,0,&name,&d,&len))
    goto bad_archive;

  if (len!=1000 || strcmp(name,"info")) goto bad_archive;
  memcpy(info.pages, d, len);
  free(d);
  file_name= fn;
  gfx_set_text(Wstatus, fn);

  create_writer_thread(par);
  return ;

bad_archive:
  gfx_ask("Error", "!OK", "bad archive %s",fn);
  far_close(&par);
}

static void clear_list()
{
  gfx_canvas_fill_rectangle(Wpixlist, 0xFFffff, 0,0, 10000, 10000);
  n_pages= 0;
  selpage= -1;
}

static void draw_sel(int idx, int color)
{
  int x,y,w,h,i;
  x= pages[idx].x;
  y= pages[idx].y;
  w= pages[idx].iw;
  h= pages[idx].ih;

  for(i=0;i<10;i++)
  {
    x--; y--; w+= 2; h+=2;
    if (i>4) gfx_canvas_draw_rectangle(Wpixlist,color, x,y,w,h);
  }
}

static void show_missing()
{
  char buf[200];
  int odd,even;
  compute_page_numbers(&odd,&even);
  sprintf(buf,"Scan start odd: %d even: %d",odd,even);
  gfx_set_text(Wpages,buf);
}


int main(int argc,char **argv)
{
  int i;
  gfx_event_t event;
  int q;
  page_t *P;

  read_config();  
  gfx_init(NULL, 1024, 900, "Book Scanner");
  gfx_SCREEN_0();
  gfx_set_state(Wrotate,1);
  gfx_set_state(Wcut,1);
  gfx_set_state(Wedge,1);
  gfx_set_state(Wmask,1);

  master_mbox= mbox_new();

  create_worker_thread();

  q= 0;
  while(!q) 
  {
  if (!gfx_event(NULL, &event)) continue;
again:
  switch(event.type)
  {
  case GFXEV(Wpixlist,BUTTON_PRESS):
    if (event.button!=1) continue;
    for(i=0;i<n_pages;i++)
      if (event.x >= pages[i].x &&
          event.x < (pages[i].x+pages[i].iw) && 
          event.y >= pages[i].y &&
          event.y < (pages[i].y+pages[i].ih))
        break;
    if (i==n_pages || i==selpage) continue;
    if (selpage!=-1) draw_sel(selpage, 0xffFFff);
    draw_sel(i, 0x00);
    selpage= i;
    break;
  case Wdelimg:
    if (selpage<0 || selpage>=n_pages) break;
    P= pages+selpage;
    if (!P->data) break;
    free(P->data);
    P->data= NULL;
    if (P->id) free(P->id); 
    P->id= NULL;
    gfx_canvas_fill_rectangle(Wpixlist, 0xFF0000, P->x,P->y, P->iw, P->ih);
    break;
  case GFXEV_QUIT: q= 1; break;
  case Wsettings: show_settings_window(); break;
  case Wcreate: 
    create_file();
    clear_list();
    show_missing();
    break;
  case Wopen:
    gfx_set_text(Wstatus,"");
    load_file();
    clear_list();
    show_missing();
    break;
  case Wreverse: scan_batch('B'); break;
  case Wnextbatch: scan_batch('F'); break;
  case Wwrite: 
    write_pages();
    clear_list();
    show_missing();
    break;
  case GFXEV_KEY_PRESS:
    if (event.key==GFX_K_PRT && toupper(event.string[0])=='F')
    { event.type= Wnextbatch; goto again; }
    if (event.key==GFX_K_PRT && toupper(event.string[0])=='B')
    { event.type= Wreverse; goto again; }
    if (event.key==GFX_K_PRT && toupper(event.string[0])=='W')
    { event.type= Wwrite; goto again; }
    break;
  }
  }

  if (par) write_index(info.pages);
  finish_writer_thread();
  if (par) far_close(&par);

  mbox_put(worker_mbox, worker_QUIT());
  while(!worker_done) master_dispatch(mbox_get(master_mbox));
  join_worker_thread();

  write_config();
  gfx_leave_screen();
  gfx_close();
  return 0;
}

#include "master.c"
