#include <math.h>
#include <stdlib.h>
#include "ops.h"
#include "rotate_image.h"


typedef struct { int x,y; } point_t;



#define unt  unsigned int
#define uhar unsigned char

int sub_image(uhar* iD, int iW, int iH,
                     int left, int right, int top, int bot,
                     uhar** Rd, unt *Rw, unt *Rh)
{
  uhar *nD;
  int x,y,nW,nH;

  nW= iW-(left+right);
  nH= iH-(bot+top);

  if (!iD || nW<=0 || nH<=0) return 1;
  nD= malloc(nW*nH);
  for(y=0;y<nH;y++)
    for(x=0;x<nW;x++)
      nD[x+nW*y]= iD[(x+left)+(y+top)*iW];
  *Rd= nD;
  *Rw= nW;
  *Rh= nH;
  return 0;
}

static int linedir(point_t q, point_t p0, point_t p1,int sense)
{
  point_t t;
  double dx, dy,m,px,qx;
  if (p0.y>p1.y) { t= p0; p0= p1; p1= t; }
  if (p0.y==p1.y) return 'C';

  dy= p1.y - p0.y;
  dx= p1.x - p0.x;
  m= dx/dy;

  px= p0.x + (q.y-p0.y)*m;
  qx= q.x;
  if (px==qx) return 'C';
  px*= sense; qx*= sense;
  if (px<qx) return 'R';
  return 'L';
}

static int free_points(point_t *P,int ret)
{
  if (P) free(P);
  return ret;
}

static int make_points(unsigned char *D, int W,int H, int thr,
                       int *Rnpoints, point_t **Rpoints, 
                       int sx, int ex, int dx)
{
  point_t *points;
  int npoints;
  int x,y;

  points= malloc(sizeof(points[0])*H);
  npoints= 0;
  for(y=1;y<H-1;y++)
    for(x=sx;x!=ex;x+=dx)
     if (D[x+y*W]>thr)
     {
       points[npoints].x= x;
       points[npoints].y= y; 
       npoints++; 
       break;
     }
  if (!npoints) return free_points(points,1);
  *Rnpoints= npoints;
  *Rpoints= points;
  return 0;
}


int black_border
     (uhar *D, unt W, unt H,
      int thr,
      unt *Rleft, unt *Rright, 
      unt *Rtop, unt *Rbot)
{
  int top,lef,rig,bot;
  int x,y;

  for(top=0;top<H;top++)     for(x=0;x<W;x++) if (D[x+top*W]>thr) goto L1;
L1:
  for(bot=0;bot+top<H;bot++) for(x=0;x<W;x++) if (D[x+(H-1-bot)*W]>thr) goto L2;
L2:
  for(lef=0;lef<W;lef++)     for(y=0;y<H;y++) if (D[lef+y*W]>thr) goto L3;
L3:
  for(rig=0;rig+lef<W;rig++) for(y=0;y<H;y++) if (D[(W-1-rig)+y*W]>thr) goto L4;
L4:

  if (lef+rig>=W) { *Rleft= 0; *Rright= W-1; }
            else  { *Rleft= lef; *Rright= rig; }
  if (top+bot>=H) { *Rtop= 0; *Rbot= H-1; }
            else  { *Rtop= top; *Rbot= bot; }

  return 0;
}

static int remove_black_border
    (unsigned char *D, unsigned int W, unsigned int H,
     int thr, 
     unsigned char **Rd, unsigned int *Rw, unsigned int *Rh)
{
  unsigned int left,right,top,bottom;
  black_border(D,W,H,thr,&left,&right,&top,&bottom);
  if (left || right || top || bottom)
    return sub_image(D,W,H,left,right,top,bottom,Rd,Rw,Rh);
  return 1;
}

static int compy;

static int comp_po(const void *A, const void *B)
{
  const point_t *a= A;
  const point_t *b= B;
  return abs(b->y-compy) - abs(a->y-compy);
}

static void sort_by_deltay(int npoints, point_t *points)
{
  compy= points[0].y;
  qsort(points+1, npoints-1, sizeof(points[0]), comp_po);
}

static int find_best_line
  (int npoints, point_t *points, int sense, point_t *Rp,int height)
{
   int i;
   int pivot;
   point_t Z;
   int found;
   int mid;

   mid= height / 2;
   pivot= 0;
   for(i=1;i<npoints;i++)
   {
     int R;
     R= points[i].x*sense - points[pivot].x*sense;
     if (R>0) continue;
     if (R<0 || abs(points[pivot].y-mid) > abs(points[i].y-mid))
       pivot= i;
   }
   Z= points[0];
   points[0]= points[pivot];
   points[pivot]= Z;

   found= 0;
   sort_by_deltay(npoints,points);
   for(i=1;i<npoints;i++)
   {
     if (abs(points[i].x-points[0].x)>=abs(points[i].y-points[0].y)) continue;
     if (!found) { found=1; Z= points[i];
// printf("found first line %d,%d -> %d,%d\n", points[0].x, points[0].y,
        // Z.x,Z.y); 
          continue; }
     if (linedir(points[i], points[0], Z, sense)=='L')
     {
       Z= points[i];
// printf("found better line %d,%d -> %d,%d\n", points[0].x, points[0].y,
        // Z.x,Z.y); 
     }
   }
   *Rp= Z;
   return !found;
}


static int find_left_edge
   (unsigned char *D, int W,int H, int thr, double *Rangle)
{
  point_t *points;
  int npoints;
  int dx,dy;
  point_t B;

  if (make_points(D,W,H, thr, &npoints, &points, 0,W,1)) return 1;

  if (find_best_line(npoints, points, 1, &B,H)) return free_points(points,1);

  dx= B.x-points[0].x;
  dy= B.y-points[0].y;
  if (dy<0)
    *Rangle= atan2(dx, -dy);
  else
    *Rangle= -atan2(dx, dy);

  return free_points(points,0);
}

static int find_right_edge
   (unsigned char *D, int W,int H, int thr, double *Rangle)
{
  point_t *points;
  int npoints;
  int dx,dy;
  point_t B;

  if (make_points(D,W,H, thr, &npoints, &points, W-1,-1,-1)) return 1;

  if (find_best_line(npoints, points, -1, &B,H)) return free_points(points,1);

  dx= B.x-points[0].x;
  dy= B.y-points[0].y;
  if (dy<0)
    *Rangle= -atan2(-dx, -dy);
  else
    *Rangle= atan2(-dx, dy);

  return free_points(points,0);
}


static double sgn(double x) { if (x<0.0) return -1; return 1; }

int process_image(uhar *iD, unt iW,unt iH,
                   uhar **Rd, unt *Rw, unt *Rh,
                   unt cutleft, unt cutright, int thr,int dir)
{
  unt cW, cH; uhar *cD= NULL;  // cut
  unt fW, fH; uhar *fD= NULL;  // final
  unt kW, kH; uhar *kD= NULL;  // 
  int R;
  double angle;
  double total;


  if ((!cutleft && !cutright) ||
       sub_image(iD, iW, iH, cutleft, cutright, 0,0, &cD, &cW,&cH))
    { cD= iD; cW= iW; cH= iH; }

  if (!remove_black_border(cD, cW, cH, thr, &fD, &fW, &fH))
  {
    if (cD!=iD) free(cD);
    cD= fD; cW= fW; cH= fH;
  }

  kD= cD; kW= cW; kH= cH;

  total= 0.0;
  while(1)
  {
    unt rW, rH; uhar *rD;  // rotated
    if (dir=='L') R= find_left_edge(cD, cW, cH, thr, &angle);
           else   R= find_right_edge(cD, cW, cH, thr, &angle);

    //printf("angle= %f (%d)\n",angle, (int) (angle/M_PI*180));

    if (R || angle==0.0 || (total!=0.0 &&sgn(angle)!=sgn(total))) break;
    rotate_image_double_grayr(cD, cW, cH, -angle, &rD, &rW, &rH);
    if (cD!=kD) free(cD);
    cD= rD; cW= rW; cH= rH;
    total+= angle;
  }
//printf("total= %f (%d)\n", total, (int) (total/M_PI*180));
  if (cD!=kD) free(cD);

  if (total!=0)
  {
    unt rW, rH; uhar *rD;  // rotated
    rotate_image_double_grayr(kD, kW, kH, -total, &rD, &rW, &rH);
    if (kD!=iD) free(kD);
    cD= rD; cW= rW; cH= rH;
  }
  else
  {
    cD= kD; cW= kW; cH= kH;
  }

  if (!remove_black_border(cD, cW, cH, thr, &fD, &fW, &fH))
  {
    if (cD!=iD) free(cD);
    cD= fD; cW= fW; cH= fH;
  }

  *Rd= cD; *Rw= cW; *Rh= cH;
  return cD==iD;
}

void apply_mask(unsigned char *D, int W,int H, int thr)
{
  int siz,i;
  siz= W*H;
  for(i=0;i<siz;i++) if (D[i]<=thr) D[i]= 0x00;
                               else D[i]= 0xff;
}

int edge_fill(unsigned char *D, unsigned int W, unsigned int H,int thr)
{
  int x,y,z,m;
    // left to right
  for(y=0;y<H;y++) 
  {
    for(x=0;x<W;x++) if (D[x+y*W]>thr) break;
    if (x==0 || x==W) continue;
    for(m=0;m<x;m++) D[m+y*W]= 0xff;
  }

  for(y=0;y<H;y++) 
  {
    for(x=W-1;x>=0;x--) if (D[x+y*W]>thr) break;
    if (x<=0 || x==W-1) continue;
    for(m=W-1;m>x;m--) D[m+y*W]= 0xff;
  }
  return 0;
}

