#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include "gfx.h"
#include "config.h"
#include "ops.h"
#include "resize_image.h"
#include "main.h"

  //     iD,iW, iH:  original scanned image
  // diD, diW, diH:  displayed image
static unsigned int iW, iH, diW, diH;
static unsigned char *iD, *diD;

  // size of the actual resized image on screen
  // used for the zooming stuff
static unsigned int riW, riH;

  // main display image
static unsigned int dW, dH;
static unsigned int *buD;

  // zoomed image
static unsigned int *zD, zW, zH;
static int zX, zY;

static void paint_background()
{
  int x,y;
  if (!buD)
    buD= malloc(dW*dH*sizeof(unsigned int));
  for(y=0;y<dH;y++)
   for(x=0;x<dW;x++)
     buD[x+y*dW]= ((x%5)/2==(y%5)/2) ? 0xccccff : 0xffffff;
}

static void put_image()
{
  gfx_pixgrid_putsimple(WipaPIX,buD);
}


static void get_config_int(int widget, int avail, int value)
{
  char buf[20];
  sprintf(buf,"%d",value);
  if (!avail) buf[0]= 0;
  gfx_set_text(widget, buf);
}

static int get_int_value(int widget, int *Rv)
{
  char *s;
  int k;
  s= gfx_get_text(widget);
  if (!s || !s[0]) return 1;
  k= atoi(s);
  if (k<0) return 1;
  *Rv= k;
  return 0;
}

static void set_config_int(int widget, char *G, int *V)
{
  if (get_int_value(widget,V)) *G= 0;
                          else *G= 1;
}

static void put_bits(unsigned char *tD,unsigned int tW, unsigned int tH)
{
  int x,y;
  for(y=0;y<tH && y<dH;y++)
   for(x=0;x<tW && x<dW;x++)
   {
     unsigned int P= tD[x+y*tW];
     buD[x+y*dW]= P | (P<<8) | (P<<16); 
   }
  put_image();
}

static void fix_zoom_coord()
{
  if (zX+zW>diW) zX= diW-zW; 
  if (zX<0) zX= 0;
  if (zY+zH>diH) zY= diH-zH;
  if (zY<0) zY= 0;
}

static void show_zoomed()
{
  int x,y;
  unsigned int P;
  if (!zD)
    zD= malloc(zW*zH*sizeof(unsigned int));
  for(y=0;y<zH;y++)
    for(x=0;x<zW;x++)
     zD[x+y*zW]= ((x%5)+(y%5))%2 ?  0xcc0033 : 0xffee00;
  if (!diD) goto done;
  fix_zoom_coord();
  for(y=0;y<zH && (y+zY)<diH;y++)
    for(x=0;x<zW && (x+zX)<diW;x++)
    {
      P= diD[(x+zX)+(y+zY)*diW];
      zD[x+y*zW]=  P | (P<<8) | (P<<16);
    }
done:
  gfx_pixgrid_putsimple(WipaZOOM, zD);
}

static void show_image(unsigned char *tD, unsigned int tW, unsigned int tH)
{
  unsigned char *rD; unsigned int rW, rH;

  paint_background();

  if (diD && diD!=tD)
  {
    if (diD!=iD) free(diD);
  }

  diD= tD; diW= tW; diH= tH; 

  if (!tD) { put_image(); show_zoomed(); return ; }
  
  if ((tW<=dW && tH<=dH) || 
          resize_image_g8_fixed(tD, tW, tH, dW, dH, &rD, &rW, &rH))
    { put_bits(tD, tW, tH); riW= tW; riH= tH; show_zoomed(); return ; }

  put_bits(rD, rW, rH); 
  riW= rW; riH= rH;
  free(rD);
  show_zoomed();
}

static void change_zoom(int x,int y)
{
  if (x<0 || y<0 || x>=diW || y>=diH || !diD || !riW || !riH) return ;
  zX= x*iW/riW-zW/2;
  zY= y*iH/riH-zH/2;
  show_zoomed();
}

void show_image_params()
{
  unsigned int left,right,top,bot;
  int mask,thr;
  char **iL;
  unsigned char *tD;
  unsigned int tW, tH;
  int dummy;
  int q,cp;

  gfx_event_t event;
  gfx_SCRimgparam();
  gfx_pixgrid_attribs(WipaPIX, &dummy, &dW, &dH);
  gfx_pixgrid_attribs(WipaZOOM, &dummy, &zW, &zH);
  get_config_int(WipaCLEFT, CONFIG.G_cutleft, CONFIG.cutleft);
  get_config_int(WipaCRIGHT, CONFIG.G_cutright, CONFIG.cutright);
  get_config_int(WipaTHR, CONFIG.G_threshold, CONFIG.threshold);
  get_config_int(WipaMASK, CONFIG.G_mask, CONFIG.mask);
  if (iD) { free(iD); iD= NULL; }
  if (diD) { free(diD); diD= NULL; } 
  paint_background();
  show_zoomed();
  put_image();
  q= 0;
  while(!q)
  {
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case WipaOK: q= 1; cp= 1; break;
    case WipaCANCEL: case GFXEV_QUIT: q=1; cp= 0; break;
    case WipaSCAN:
      iL= scan_images();
      if (!iL || !iL[0]) continue;
      if (iD) free(iD);
      if (diD!=iD && diD) free(diD);
      iD= diD= NULL;
      if (read_image(iL[0]+1,&iD, &iW, &iH))
         show_image(NULL, 0,0);
      else
         show_image(iD, iW,iH);
      remove_images(iL);
      break;
    case WipaRESET:
      if (iD) show_image(iD, iW,iH);
      break;
    case WipaAPPLYcleft:
      if (!diD) break;
      if (get_int_value(WipaCLEFT,(int*) &left)) continue;
      if (sub_image(diD, diW, diH, left, 0,0,0,&tD, &tW, &tH)) continue;
      show_image(tD, tW, tH);
      break;
    case WipaAPPLYcright:
      if (!diD) break;
      if (get_int_value(WipaCRIGHT,(int*) &right)) continue;
      if (sub_image(diD, diW, diH, 0, right,0,0,&tD, &tW, &tH)) continue;
      show_image(tD, tW, tH);
      break;
    case WipaAPPLYthr:
      if (!diD) break;
      if (get_int_value(WipaTHR,(int*)&thr)) continue;
      black_border(diD, diW, diH, thr, &left, &right, &top, &bot);
      if (sub_image(diD, diW, diH, left, right,top,bot,&tD, &tW, &tH)) 
                 continue;
      show_image(tD, tW, tH);
      break;
    case WipaPIX+GFXEV_BUTTON_PRESS:
      if (event.button!=1) continue;
      change_zoom(event.x,event.y);
      break;
    case WipaAPPLYmask:
      if (!iD) break;
      if (get_int_value(WipaMASK,&mask)) break;
      tD= malloc(iW*iH);
      memcpy(tD, iD, iW*iH);
      apply_mask(tD, iW, iH, mask);
      show_image(tD, iW, iH);
      break;
    }
  }
  if (cp)
  {
    set_config_int(WipaCLEFT, &CONFIG.G_cutleft, &CONFIG.cutleft);
    set_config_int(WipaCRIGHT, &CONFIG.G_cutright, &CONFIG.cutright);
    set_config_int(WipaTHR, &CONFIG.G_threshold, &CONFIG.threshold);
    set_config_int(WipaMASK, &CONFIG.G_mask, &CONFIG.mask);
  }
  if (diD && diD!=iD) free(diD);
  if (iD) free(iD);
  if (buD) free(buD);
  if (zD) free(zD);
  zD= NULL;
  buD= NULL; iD= diD= NULL;
  gfx_leave_screen();
}
