#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h> 
#include <string.h>
#include <unistd.h>
#include "proc.h"
#include "mbox.h"
#include "gfx.h"
#include "far.h"
#include "cmdline.h"
#include "config.h"
#include "util.h"
#include "main.h"
#include "readpgm.h"

static void *iptr(void* value, int dum) { return value; }


static char **collect_scanned_files()
{
  char **L;
  int N,i;
  static char *imgext[]= { ".pgm", ".pnm", ".ppm", NULL };
  L= list_files(CONFIG.tempdir);
  for(N=0;L[N];N++) ;
  file_types(CONFIG.tempdir, L);
  for(i=0;i<N;i++)
    if (L[i][0]!='F') { L[i]= L[N-1]; N--; i--; }
  N= filter_files_byhid(N, L, 0);
  N= filter_files_byext(N, L, imgext);
  sort_files(N, L);
  L[N]= NULL;
  if (N==0) { free(L); return NULL; }
  return L;
}

static int message(char *msg, int dum)
{
  gfx_insert(WscaDISP,-1, msg, NULL, NULL);
  return 0;
}

static char **scan_params()
{
  char **cmd; int cmde;

  if (!CONFIG.G_tempdir)
    return iptr(NULL, message("please set temporary directory",0));
 
  mkdir_p(CONFIG.tempdir);
  if (!CONFIG.G_scancmd)
    return iptr(NULL, message("please set scan command",0));

  cmd= parse_cmd_line(CONFIG.scancmd, &cmde);
  int i;
  for(i=0;cmd[i];i++)
  {
   if (cmd[i][0]=='\'')
   {
     int j;
     cmd[i]++;
     for(j=0;cmd[i][j];j++) if (cmd[i][j]=='\'') { cmd[i][j]= 0; break; }
   }
  }
  if (!cmd)
    return iptr(NULL, message("bad scan command line",0));
  return cmd;
}


static int get_output(int fd, char *buf, int *len)
{
  int R;
  char c;
  R= read(fd, &c,1);
  if (R<0) return 1;
  if (R==0) return 0;
  if (c=='\r') return 0;
  if (c=='\n' || *len==1000) {  message(buf,buf[*len]= 0); *len= 0; }
  else { buf[*len]= c; (*len)++; }
  return 0;
}

char** scan_images()
{
  gfx_event_t event;
  char **cmd; 
  int child;
  char buf[2][2000];
  int blen[2];
  int fd[2];
  int close_scr;

  close_scr= 0;
  gfx_SCRscan();

  if (!(cmd=scan_params())) goto finish;
  
  child= proc_start(cmd[0],cmd,PROC_OUT|PROC_ERR|PROC_PTY);
  if (child<0) goto finish;

  fd[0]= proc_output(child);
  fd[1]= proc_error(child);

  if (fd[1]>=0) gfx_monitor_file(fd[1], 1, 1,0,1);
  if (fd[0]>=0) gfx_monitor_file(fd[0], 0, 1,0,1);

  blen[0]= blen[1]= 0;

  while(1) 
  {
  unsigned int tim;
  tim= 200;
  if (!gfx_event(&tim,&event))
  {
    if (proc_islive(child)) continue;
    goto rip_child;
  }
  switch(event.type)
  {
  case GFXEV_READY_READ:
    if (get_output(fd[event.key], buf[event.key], &blen[event.key]) &&
        !proc_islive(child)) goto rip_child;
    break;
  case GFXEV_READY_EXCEPT:
    gfx_stop_monitor_file(fd[event.key], event.key);
    fd[event.key]= -1;
    if (fd[0]==fd[1])
    {
      proc_end(child, NULL);
      goto finish;
    }
    break;
  case GFXEV_QUIT:
  case WscaCLOSE:
    close_scr= 1;
    proc_kill(child, SIGKILL);
    goto rip_child;
  }
  }
rip_child:
  gfx_stop_monitor_file(fd[0],0);
  gfx_stop_monitor_file(fd[1],1);
  proc_end(child, NULL);
  if (!close_scr) goto finish;
  gfx_leave_screen();
  goto collect_files;

finish:
  while(1) if (gfx_event(NULL,&event)) switch(event.type)
  { case GFXEV_QUIT: case WscaCLOSE: gfx_leave_screen(); goto collect_files ; }

collect_files:
  return collect_scanned_files();
}

void remove_image(char *fn)
{
  char buf[2000];
  if (!fn) return ;
  if (!CONFIG.G_tempdir || !CONFIG.tempdir || !CONFIG.tempdir[0]) return ;
  snprintf(buf,sizeof(buf),"%s/%s",CONFIG.tempdir, fn);
  buf[sizeof(buf)-1]= 0;
  unlink(buf);
}

void remove_images(char **L)
{
  int i;
  if (!L) return ;
  if (!CONFIG.G_tempdir || !CONFIG.tempdir || !CONFIG.tempdir[0]) return ;
  for(i=0;L[i];i++)
  {
    char buf[2000];
    snprintf(buf,sizeof(buf),"%s/%s",CONFIG.tempdir, L[i]+1);
    buf[sizeof(buf)-1]= 0;
    unlink(buf);
  }
}

int read_image(char *fn,unsigned char **Rd,unsigned int *Rw,unsigned int *Rh)
{
  char buf[2000];
  char *e;
  if (!fn) return 1;
  if (!CONFIG.G_tempdir || !CONFIG.tempdir || !CONFIG.tempdir[0]) return 1;
  snprintf(buf,sizeof(buf),"%s/%s",CONFIG.tempdir, fn);
  buf[sizeof(buf)-1]= 0;
  e= pgm_read_file(buf,Rd,Rw,Rh); 
  if (e) return 1;
  return 0;
}
